#include "backup.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "config.h"
#include "command.h"
#include "iv.h"
#include "mess.h"
#include "sha1.h"

char *backup_dir;
char *backup_single = "/";
int backup_depth = -1;
int skip_phpXXXXXX;

extern buf from_server;

static char curdir[FILENAME_MAX + 1];
static int cursize;

static void push(char *s)
{
  int l = strlen(s);
  /* ensure that curdir starts with a '/' */
  if (s[0] != '/' && cursize == 0) {
    curdir[0] = '/';
    cursize = 1;
  }
  if (cursize + l > FILENAME_MAX) {
    fprintf(stderr, "PUSH ERROR\n");
    exit(1);
  }
  memcpy(curdir+cursize, s, l+1);
  cursize += l;
  /* ensure that curdir ends with a '/' */
  if (s[l-1] != '/') {
    if (cursize + 1 > FILENAME_MAX) {
      fprintf(stderr, "PUSH ERROR\n");
      exit(1);
    }
    curdir[cursize] = '/';
    cursize++;
  }
}

static void pop(void)
{
  cursize--;
  while (1) {
    cursize--;
    if (cursize < 0) break;
    if (curdir[cursize] == '/') break;
  }
  cursize++;
  curdir[cursize] = 0;
}

void makedir(char *dir)
{
  struct stat b;
  char s[FILENAME_MAX+1];
  if (strlen(dir) + strlen(backup_dir) > FILENAME_MAX) {
    fprintf(stderr, "makedir ERROR\n");
    exit(1);
  }
  sprintf(s, "%s%s", backup_dir, dir);
  if (stat(s, &b) == -1) {
    if(mkdir(s, 0755) == -1) {
      perror(s);
      exit(1);
    }
    return;
  }
  if (!S_ISDIR(b.st_mode)) {
    fprintf(stderr, "%s: not a directory\n", s);
    exit(1);
  }
}

void getfile(char *f, int size)
{
  FILE *ff;
  struct stat b;
  char local[FILENAME_MAX+1];
  char remote[FILENAME_MAX+1];
  int curpos = 0;
  int try;
  int second_pass = 0;
  SHA1Context ctx;
  unsigned char buf[4096];
  unsigned char sha_from_server[20];
  unsigned char local_sha[20];
  int i;

  if (size == 0) {
    printf("WARNING: size 0 for %s, not retrieving\n", f);
    return;
  }

  if (skip_phpXXXXXX && strlen(f) == 9 &&
      f[0] == 'p' && f[1] == 'h' && f[2] == 'p') {
    printf("WARNING: skipping %s\n", f);
    return;
  }

  if (strlen(backup_dir) + strlen(curdir) + strlen(f) > FILENAME_MAX) {
    fprintf(stderr, "getfile ERROR\n");
    exit(1);
  }
  sprintf(local, "%s%s%s", backup_dir, curdir, f);
  sprintf(remote, "%s%s", curdir, f);

  printf("GETTING %s (%s) %d\n", remote, local, size);

  /* if the file is there, maybe it's a continue */
  if (stat(local, &b) == 0)
    curpos = b.st_size;

  /* but if the local size is bigger than remote we are sure the local
   * copy is wrong
   */
  if (size < curpos) {
    printf("WARNING: local file's size bigger than remote\n");
    unlink(local);
    curpos = 0;
  }

try_again:
  while (curpos != size) {
    int todo = size - curpos;
    if (todo > GET_BLOCK_SIZE) todo = GET_BLOCK_SIZE;
    try = 0;
one_more_gt:
    printf("GT %d %d\n", curpos, todo);
    gt(&from_server, local, remote, curpos, todo);
    if (from_server.size == 0) {
      if (try == 3) {
        fprintf(stderr, "TRY TOO MUCH\n");
        exit(1);
      }
      synchronize_iv(&from_server);
      try++;
      printf("TRY AGAIN!\n");
      goto one_more_gt;
    }
    curpos += todo;
  }

  /* local and remote copies must have same SHA1.
   * If not, delete local copy and try again.
   * If the SHA1 differs the second time then something
   * weird happened. Just give up. (Except for special
   * case "iv.php" which may be part of phpoor and always
   * changes, in which case we skip it.) (Maybe we
   * should insure that it's the right iv.php? Or
   * a command line option?)
   */
  try = 0;
one_more_sh:
  sh(&from_server, remote, (char *)sha_from_server);
  if (from_server.size == 0) {
    if (try == 3) {
      fprintf(stderr, "TRY TOO MUCH 2\n");
      exit(1);
    }
    synchronize_iv(&from_server);
    try++;
    printf("TRY AGAIN!\n");
    goto one_more_sh;
  }

  SHA1Reset(&ctx);
  if (stat(local, &b) != 0) { perror(local); exit(1); }
  curpos = b.st_size;
  ff = fopen(local, "rb"); if (!f) { perror(local); exit(1); }
  while (curpos) {
    int todo = curpos;
    if (todo > 4096) todo = 4096;
    if (fread(buf, todo, 1, ff) != 1) { perror(local); exit(1); }
    curpos -= todo;
    SHA1Input(&ctx, buf, todo);
  }
  fclose(ff);
  SHA1Result(&ctx, local_sha);

  for (i = 0; i < 20; i++)
    if (sha_from_server[i] != local_sha[i]) {
      if (second_pass) {
        fprintf(stderr, "FATAL ERROR GUY\n");
        exit(1);
      }
      if (!strcmp(remote+strlen(remote)-6, "iv.php")) {
        printf("WARNING: bad SHA1 for %s, but we keep the current copy!\n",
               remote);
        return;
      }
      second_pass = 1;
      curpos = 0;
      unlink(local);
      printf("WARNING: bad SHA1, try again with %s\n", remote);
      goto try_again;
    }
}

void go(char *dir, int depth)
{
  int loop = 0;
  int i;
  int size;
  int len;
  list in_list;

  if (backup_depth != -1 && depth > backup_depth) return;
  depth++;

  memset(&in_list, 0, sizeof(list));

  push(dir);
  makedir(curdir);

  printf("DOING %s\n", curdir);

more:
  ls(&from_server, curdir, &in_list);
  if (from_server.size == 0) {
    synchronize_iv(&from_server);
    loop++;
    printf("LS TRY AGAIN!\n");
    if (loop < 3) goto more;
    fprintf(stderr, "error somewhere in ls\n");
    exit(1);
  }

  for (i = 0; i < list_size(&in_list); i++) {
    char *cur = in_list.buf + in_list.data[i];
    if (!strcmp(cur, "./")) continue;
    if (!strcmp(cur, "../")) continue;
    len = strlen(cur);
    sscanf(cur + len + 1, "%d", &size);
    if (cur[len-1] == '/') printf("DIR ");
    printf("%s %d\n", cur, size);
    if (cur[len-1] == '/') { go(cur, depth); continue; }
    getfile(cur, size);
  }

  pop();

  if (in_list.data) free(in_list.data);
  if (in_list.buf) free(in_list.buf);
}

void backup_main(void)
{
  go(backup_single, 0);
  exit(0);
}

//#define WITHMESS

#ifdef WITHMESS
static void lmess(const char *l, ...)
{
  va_list ap;
  va_start(ap, l);
  vfprintf(stderr, l, ap);
  va_end(ap);
  fflush(stderr);
}

static void lfmess(const void *b, int l)
{
  fwrite(b, l, 1, stderr);
  fflush(stderr);
}
#endif

void backup_mess(void)
{
#ifdef WITHMESS
  mess = lmess;
  fmess = lfmess;
#endif
}
