#include "com.h"
#include "config.h"
#include "request.h"

#include "mess.h"
#include "statistics.h"

#include <sys/types.h>
#ifndef WINDOZ
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/select.h>
#endif
#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef WINDOZ
#include <winsock2.h>
#endif

static unsigned char reply[4096];
static int reply_size;
static int reply_pos;

char *server;
int port;
char *server_url;

static void get_buffer(int fd)
{
  int d;
#ifndef WINDOZ
  d = read(fd, reply, 4096);
#else
  d = recv(fd, reply, 4096, 0);
#endif
  if (d <= 0) return;
  reply_pos = 0;
  reply_size = d;

  request_reply_size += d;

  mess("**** RAW BUFFER IN BEGIN size=%d\n", d);
  fmess(reply, reply_size);
  mess("\n**** RAW BUFFER IN END\n");
}

/* return -1 if error */
static int get_char(int fd)
{
  int ret;
  if (reply_pos == reply_size) {
    get_buffer(fd);
    if (reply_pos == reply_size) return -1;
  }
  ret = reply[reply_pos];
  reply_pos++;
  return ret;
}

static void get_line(int fd, buf *b)
{
  int c;

  buf_reset(b);

  while (1) {
    c = get_char(fd);
    /* error or no data, let's pretend it's a end of string */
    if (c == -1) { buf_add_char(b, 0); return; }
    if (c == 13) {
      c = get_char(fd);
//      if (c == -1) goto error;
      /* error or no data, let's pretend it's a end of string */
      if (c == -1) { buf_add_char(b, 0); return; }
      if (c == 10) { buf_add_char(b, 0); return; }
      buf_add_char(b, 13);
      buf_add_char(b, c);
    }
    buf_add_char(b, c);
  }
  return;
}

static buf line;
void post(buf *out)
{
  struct hostent *h;
  struct sockaddr_in addr;
  int fd;
#ifdef WINDOZ
  static int inited = 0;
#endif
  int l;
  char *s;
  int d;
  int data_size = 0; /* 0 at start, -1 if chunked */

  reply_pos = 0;
  reply_size = 0;
  buf_reset(out);

#ifdef WINDOZ
  /* they can't do like everyone else... */
  if (!inited) {
    WSADATA wsaData;
    int iResult;
    iResult = WSAStartup(MAKEWORD(2, 2), &wsaData);
    if (iResult != 0) {
        mess("WSAStartup failed: %d\n", iResult);
        return;
    }
    inited = 1;
  }
#endif

  /* let's connect */
  h = gethostbyname(server);
  if (h == 0)
    return;

  /* is this code clean ? I mean *(unsigned long *)... */
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = *(unsigned long *)h->h_addr_list[0];
  addr.sin_port = htons(port);

  fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (fd == -1)
    return;

  if (connect(fd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in)) == -1) {
    return;
  }

  request_size = r_size();

  mess("**** RAW REQUEST BEGIN\n");
  fmess(r_req(), r_size());
  mess("\n**** RAW REQUEST END\n");

  /* write request */
  s = r_req();
  l = r_size();
  while (l) {
#ifndef WINDOZ
    d = write(fd, s, l);
#else
    d = send(fd, s, l, 0);
#endif
    if (d <= 0) goto error;
    l -= d;
    s += d;
  }

  request_reply_size = 0;

  /* get reply */
  /* get HTTP headers */
  /* first line is status */
  get_line(fd, &line);
  /* should be OK, otherwise return nothing (this test might fail) */
  if (strncmp(line.buf, "HTTP/1.1 200", strlen("HTTP/1.1 200")) &&
      strncmp(line.buf, "HTTP/1.0 200", strlen("HTTP/1.0 200"))) {
    mess("HTTP reply: %s\n", line.buf);
    buf_reset(out);
    goto error;
  }

  /* other lines we don't care, just get the size (or chunked stuff) */
  while (1) {
    get_line(fd, &line);
//printf("here it's ok %d\n", data_size);
//fwrite(line.buf, line.size, 1, stdout); printf("\n");
    /* no line means error */
    if (line.size == 0) goto error;
    /* empty line means end of headers */
    if (line.size == 1) break;
    if (!strncmp(line.buf, "Content-Length:", strlen("Content-Length:"))) {
      int d;
      /* size already specified => error */
      if (data_size != 0) goto error;
      if (sscanf(line.buf, "Content-Length: %d", &d) != 1) goto error;
      /* no data? error (the server alaways returns something) */
      if (d == 0) goto error;
      data_size = d;
    } else
    if (!strncmp(line.buf, "Transfer-Encoding: chunked",
                 strlen("Transfer-Encoding: chunked"))) {
      if (data_size != 0) goto error;
      data_size = -1;
    }
  }

  /* no size/chunk stuff found, get all we can */
  if (data_size == 0) {
    while (1) {
      int c = get_char(fd);
      if (c == -1) break;
      buf_add_char(out, c);
    }
    goto end;
  }

  /* chunked? */
  if (data_size == -1) { while (1) {
    get_line(fd, &line);
    if (line.size == 0) goto error;
    if (sscanf(line.buf, "%x", &d) != 1) goto error;
    if (d == 0) goto chunked_end;
    if (d < 0) goto error;
mess("chunk size=%d\n", d);
    while (d) {
      int c = get_char(fd);
      if (c == -1) goto error;
      buf_add_char(out, c);
      d--;
    }
    get_line(fd, &line);
    if (line.size != 1) goto error;
  }
chunked_end:
  /* get trailer */
  do get_line(fd, &line); while (line.size > 1);
  if (line.size != 1) goto error;
  goto end;
  }

  /* no => read data */
  while (data_size) {
    int c = get_char(fd);
    if (c == -1) goto error;
    buf_add_char(out, c);
    data_size--;
  }

end:
  close(fd);
  return;

error:
  close(fd);
  mess("post error\n");

//fwrite(out->buf, out->size, 1, stdout);

  buf_reset(out);
}
