#include "crypto.h"

#include "mess.h"

#include "ecrypt-sync.h"
#include "sha1.h"

#include <stdio.h>

static unsigned char key[256/8*2];
static ECRYPT_ctx ctx;
static unsigned char xorkey[32];
static unsigned char k[256/8];
static unsigned char v[128/8];

/* not 100% secure, but it's better than nothing (the key is not
 * cleared in case of emergency exit, the key is always in clear
 * for all the program's life, and so on)
 */
void clear_key(void)
{
  int i;
  for (i=0; i<256/8; i++) { k[i] = 0; xorkey[i] = 0; }
}

/* return -1 if no key read, 0 if ok */
int read_key(const char *file, const char *pass, int pass_len)
{
  SHA1Context ctx;
  int i;
  FILE *f = fopen(file, "rb");
  if (!f) {
error:
    mess("%s: file not found.\n", file);
    return -1;
  }
  if (fread(key, 256/8*2, 1, f) != 1) { fclose(f); goto error; }
  fclose(f);

  SHA1Reset(&ctx);
  SHA1Input(&ctx, (const unsigned char *)pass, pass_len);
  SHA1Result(&ctx, (unsigned char *)xorkey);
  for (i=20; i<32; i++) xorkey[i] = xorkey[i-20];
  return 0;
}

void init_crypto(char *iv)
{
  char t[3];
  int i;
  int d;

  t[2] = 0;

  /* key and iv in binary */
  /* ugly? a bit... but run once, so no big issue */
  for (i = 0; i < 256/8; i++) {
    t[0] = key[i*2];
    t[1] = key[i*2+1];
    sscanf(t, "%x", &d);
    k[i] = d & 0xff;
  }
  for (i = 0; i < 64/8+8; i++) {
    t[0] = iv[i*2];
    t[1] = iv[i*2+1];
    sscanf(t, "%x", &d);
    v[i] = d & 0xff;
  }

  for (i = 0; i < 32; i++) k[i] ^= xorkey[i];

  ECRYPT_keysetup(&ctx, k, 256, 64);
  ECRYPT_ivsetup(&ctx, v);
}

void cypher(char *to, char *from, int size)
{
  ECRYPT_encrypt_bytes(&ctx, (unsigned char *)from, (unsigned char *)to, size);
}

unsigned char *get_key(void)
{
  return k;
}

/* return current iv, don't fuck up the returned value,
 * don't free it, it's a local static buffer
 */
unsigned char *get_iv(void)
{
  static unsigned char dv[16];
  U32TO8_LITTLE(dv, ctx.input[6]);
  U32TO8_LITTLE(dv+4, ctx.input[7]);
  U32TO8_LITTLE(dv+8, ctx.input[8]);
  U32TO8_LITTLE(dv+12, ctx.input[9]);
  return dv;
}
