#include "gui_command.h"

#include "mess.h"
#include "command.h"
#include "config.h"

#include "sha1.h"

extern buf from_server;

#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

/* is_continue = 1: continue getting file from what we already have
 *             = 0: start from scratch
 */
void do_get(main_gui *m, int is_continue)
{
  char *remote_filename;
  char filename[FILENAME_MAX];
  int filesize;
  int size;
  int pos;
  int totsize;
  unsigned char sha_from_server[20];
  unsigned char local_sha[20];
  unsigned char buf[4096];
  SHA1Context ctx;
  FILE *f;
  int i;
  struct stat bstat;

  remote_filename = list_get(m->remote_files, m->line);
  sscanf(remote_filename + strlen(remote_filename) + 1, "%d", &filesize);

  if (strlen(remote_filename) + strlen(m->remote_dir) + 1 > FILENAME_MAX) {
    mess("Cannot get file, filename too long: %s%s\n",
         m->remote_dir, remote_filename);
    return;
  }
  sprintf(filename, "%s%s", m->remote_dir, remote_filename);

  if (filesize == 0) {
    mess("Get %s: file's size is null, try to refresh, maybe. ",
         remote_filename);
    mess("Note also that directories cannot be retrieved.\n");
    return;
  }

  /* we delete the file if it already exists if not in continue mode */
  if (!is_continue)
    unlink(remote_filename);

  totsize = filesize;

  if (is_continue) {
    if (stat(remote_filename, &bstat) == -1) {
st_err:
      mess("Get %s: error stat file for continue transfer. ",
           remote_filename);
      return;
    }
    pos = bstat.st_size;
    filesize -= bstat.st_size;
    /* paranoid check, probably useless */
    if (filesize < 0) goto st_err;
  }

  if (!is_continue) pos = 0;

  while (filesize) {
    size = filesize;
    if (size > GET_BLOCK_SIZE) size = GET_BLOCK_SIZE;
    gt(&from_server, remote_filename, filename, pos, size);
    if (from_server.size == 0) {
      mess("Get %s: suberror in gt\n", remote_filename);
      return;
    }
    filesize -= size;
    pos += size;
    mess("Get %s: %d/%d\n", remote_filename, pos, totsize);
  }

  refresh_local_files(m);

  sh(&from_server, filename, (char *)sha_from_server);
  if (from_server.size == 0) {
    mess("Get %s: suberror in sh\n", remote_filename);
    return;
  }

  SHA1Reset(&ctx);

  f = fopen(remote_filename, "rb");
  if (!f) {
    mess("Get %s: error in opening file for SHA verification\n",
         remote_filename);
    return;
  }

  filesize = totsize;
  while (filesize) {
    size = filesize;
    if (size > 4096) size = 4096;
    if (fread(buf, size, 1, f) != 1) {
      fclose(f);
      mess("Get %s: error in reading file for SHA verification\n",
           remote_filename);
      return;
    }
    filesize -= size;
    SHA1Input(&ctx, buf, size);
  }
  fclose(f);

  SHA1Result(&ctx, local_sha);

  for (i = 0; i<20; i++)
    if (sha_from_server[i] != local_sha[i]) {
      mess("Get %s: error in SHA verification\n", remote_filename);
      return;
    }

  mess("Get %s: OK\n", remote_filename);
}

/* is_continue = 1: continue putting file from what we already have
 *             = 0: start from scratch
 */
void do_put(main_gui *m, int is_continue)
{
  char *local_filename;
  char filename[FILENAME_MAX];
  int filesize;
  struct stat bstat;
  int size;
  int pos;
  int totsize;
  unsigned char sha_from_server[20];
  unsigned char local_sha[20];
  unsigned char buf[4096];
  SHA1Context ctx;
  FILE *f;
  int i;

  local_filename = list_get(m->local_files, m->line);
  if (stat(local_filename, &bstat) == -1) {
    mess("Error stat file: %s\n", local_filename);
    return;
  }
  filesize = bstat.st_size;
  if (filesize <= 0) {
    mess("Error file: %s has size: %d\n", local_filename, filesize);
    return;
  }

  if (strlen(local_filename) + strlen(m->remote_dir) + 1 > FILENAME_MAX) {
    mess("Cannot put file, filename too long: %s%s\n",
         m->remote_dir, local_filename);
    return;
  }
  sprintf(filename, "%s%s", m->remote_dir, local_filename);

  if (!is_continue) {
    /* we delete remote file if it exists */
    for (i=0; i<m->remote_files->size; i++)
      if (!strcmp(local_filename, list_get(m->remote_files, i))) {
        df(&from_server, filename);
        break;
      }
  }

  totsize = filesize;

  if (is_continue) {
    int remotesize = -1;
    for (i=0; i<m->remote_files->size; i++)
      if (!strcmp(local_filename, list_get(m->remote_files, i))) {
        char *fn = list_get(m->remote_files, i);
        sscanf(fn + strlen(fn) + 1, "%d", &remotesize);
        break;
      }
    if (remotesize <= 0) {
sz_err:
      mess("Put %s: bad size for continue transfer: %d\n",
           local_filename, remotesize);
      return;
    }

    pos = remotesize;
    filesize -= remotesize;
    /* paranoid check, probably useless */
    if (filesize < 0) goto sz_err;
  }

  if (!is_continue)
    pos = 0;

  while (filesize) {
    size = filesize;
    if (size > PUT_BLOCK_SIZE) size = PUT_BLOCK_SIZE;
    pt(&from_server, local_filename, filename, pos, size);
    if (from_server.size == 0) {
      mess("Put %s: suberror in pt\n", local_filename);
      return;
    }
    filesize -= size;
    pos += size;
    mess("Put %s: %d/%d\n", local_filename, pos, totsize);
  }

  ls(&from_server, m->remote_dir, m->remote_files);
  refresh_remote_files(m);

  sh(&from_server, filename, (char *)sha_from_server);
  if (from_server.size == 0) {
    mess("Put %s: suberror in sh\n", local_filename);
    return;
  }

  SHA1Reset(&ctx);

  f = fopen(local_filename, "rb");
  if (!f) {
    mess("Put %s: error in opening file for SHA verification\n",
         local_filename);
    return;
  }

  filesize = totsize;
  while (filesize) {
    size = filesize;
    if (size > 4096) size = 4096;
    if (fread(buf, size, 1, f) != 1) {
      fclose(f);
      mess("Put %s: error in reading file for SHA verification\n",
           local_filename);
      return;
    }
    filesize -= size;
    SHA1Input(&ctx, buf, size);
  }
  fclose(f);

  SHA1Result(&ctx, local_sha);

  for (i = 0; i<20; i++)
    if (sha_from_server[i] != local_sha[i]) {
      mess("Put %s: error in SHA verification\n", local_filename);
      return;
    }

  mess("Put %s: OK\n", local_filename);
}

void do_verify(main_gui *m)
{
  char *local_filename;
  char filename[FILENAME_MAX];
  int filesize;
  struct stat bstat;
  int size;
  unsigned char sha_from_server[20];
  unsigned char local_sha[20];
  unsigned char buf[4096];
  SHA1Context ctx;
  FILE *f;
  int i;

  local_filename = list_get(m->local_files, m->line);
  if (stat(local_filename, &bstat) == -1) {
    mess("Error stat file: %s\n", local_filename);
    return;
  }
  filesize = bstat.st_size;
  if (filesize <= 0) {
    mess("Error file: %s has size: %d\n", local_filename, filesize);
    return;
  }

  if (strlen(local_filename) + strlen(m->remote_dir) + 1 > FILENAME_MAX) {
    mess("Cannot verify file, filename too long: %s%s\n",
         m->remote_dir, local_filename);
    return;
  }
  sprintf(filename, "%s%s", m->remote_dir, local_filename);

  sh(&from_server, filename, (char *)sha_from_server);
  if (from_server.size == 0) {
    mess("Verify %s: suberror in sh\n", local_filename);
    return;
  }

  SHA1Reset(&ctx);

  f = fopen(local_filename, "rb");
  if (!f) {
    mess("Verify %s: error in opening file for SHA verification\n",
         local_filename);
    return;
  }

  while (filesize) {
    size = filesize;
    if (size > 4096) size = 4096;
    if (fread(buf, size, 1, f) != 1) {
      fclose(f);
      mess("Verify %s: error in reading file for SHA verification\n",
           local_filename);
      return;
    }
    filesize -= size;
    SHA1Input(&ctx, buf, size);
  }
  fclose(f);

  SHA1Result(&ctx, local_sha);

  for (i = 0; i<20; i++)
    if (sha_from_server[i] != local_sha[i]) {
      mess("Verify %s: error in SHA verification\n", local_filename);
      return;
    }

  mess("Verify %s: OK\n", local_filename);
}

void do_mkdir(main_gui *m)
{
  static mkdir_widget md;
  init_mkdir_widget(&md, m, m->win);
  gtk_widget_show_all(md.win);
}

void do_del(main_gui *m)
{
  char *t;
  char z[FILENAME_MAX];
  int len;
  int l2;

  t = list_get(m->remote_files, m->line);
  len = strlen(t);
  l2 = strlen(m->remote_dir);

  if (len + l2 + 1 > FILENAME_MAX) {
    mess("Cannot delete, filename too long: %s%s\n", m->remote_dir, t);
    return;
  }
  sprintf(z, "%s%s", m->remote_dir, t);

  if (t[len-1] == '/') {
    dd(&from_server, z);
  } else {
    df(&from_server, z);
  }

  ls(&from_server, m->remote_dir, m->remote_files);
  refresh_remote_files(m);
}

void do_move(main_gui *m, char *to)
{
  char *t;
  char z[FILENAME_MAX];
  int len;
  int l2;

  t = list_get(m->remote_files, m->line);
  len = strlen(t);
  l2 = strlen(m->remote_dir);

  if (len + l2 + 1 > FILENAME_MAX) {
    mess("Cannot move file, filename too long: %s%s\n", m->remote_dir, t);
    return;
  }
  sprintf(z, "%s%s", m->remote_dir, t);

  rn(&from_server, z, to);
  ls(&from_server, m->remote_dir, m->remote_files);
  refresh_remote_files(m);
}
