#include "iv.h"

#include "crypto.h"
#include "command.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

static char iv_file[FILENAME_MAX];

void load_iv(void)
{
  char iv[128/8*2];
  FILE *f = fopen(iv_file, "rb");
  if (!f) {
    printf("Cannot open IV file: %s\n", iv_file);
    exit(1);
  }
  if (fread(iv, 32, 1, f) != 1) {
    printf("Cannot read IV file: %s\n", iv_file);
    exit(1);
  }
  if (fclose(f) != 0) {
    printf("Cannot close IV file: %s\n", iv_file);
    exit(1);
  }
  init_crypto(iv);
}

void save_iv(void)
{
  unsigned char *iv = get_iv();
  FILE *f = fopen(iv_file, "wb");
  int i;
  if (!f) {
    printf("Cannot open IV file: %s\n", iv_file);
    exit(1);
  }
  for (i=0; i<16; i++)
    if (fprintf(f, "%2.2x", iv[i]) == -1) {
      printf("Cannot write IV file: %s\n", iv_file);
      exit(1);
    }
  if (fclose(f) != 0) {
    printf("Cannot close IV file: %s\n", iv_file);
    exit(1);
  }
}

void set_iv_file(char *f)
{
  int l = strlen(f);
  char cwd[FILENAME_MAX];
  if (getcwd(cwd, FILENAME_MAX-1) == NULL) {
    printf("Error with getcwd\n");
    exit(1);
  }
  if (l + strlen(cwd) + 1 > FILENAME_MAX-1) {
    printf("Bad IV file: %s\n", f);
    exit(1);
  }
  sprintf(iv_file, "%s/%s", cwd, f);
}

/* return -1 if a<b, 1 if a>b, 0 if a=b */
int compare_iv(unsigned char *a, unsigned char *b)
{
  /* 6: 0, 7:1, 8:2, 9:3 */
  /* inc 8, 9, 6, 7 */
  int order[] = {
    /* 7 is bigger */
    1 * 4 + 3, 1 * 4 + 2, 1 * 4 + 1, 1 * 4 + 0,
    /* then 6 */
    0 * 4 + 3, 0 * 4 + 2, 0 * 4 + 1, 0 * 4 + 0,
    /* then 9 */
    3 * 4 + 3, 3 * 4 + 2, 3 * 4 + 1, 3 * 4 + 0,
    /* finally 8 */
    2 * 4 + 3, 2 * 4 + 2, 2 * 4 + 1, 2 * 4 + 0,
  };
  int i;
  for (i = 0; i < 16; i++) {
    if (a[order[i]] < b[order[i]]) return -1;
    if (a[order[i]] > b[order[i]]) return 1;
  }
  return 0;
}

void synchronize_iv(buf *out)
{
  unsigned char remote_iv[16];
  unsigned char *local_iv;
  int i;
  char t[3];
  extern int must_synchronize_iv;
  int ret;

  must_synchronize_iv = 0;

  load_iv();
  local_iv = get_iv();

  iv(out);
  if (out->size == 0) goto error;

  t[2] = 0;
  for (i=0; i<16; i++) {
    int v;
    t[0] = out->buf[i*2];
    t[1] = out->buf[i*2+1];
    sscanf(t, "%x", &v);
    remote_iv[i] = v&0xff;
  }

  ret = compare_iv(local_iv, remote_iv);

  /* ret <= 0 means local_iv <= remote_iv, so we must use remote_iv */
  if (ret <= 0) {
    init_crypto(out->buf);
    save_iv();
  } else {
    /* load_iv did the init_crypto */
    iw(out, (char *)local_iv);
    /* iw failed? synchronizing must be done again */
    if (out->size == 0) goto error;
  }

  must_synchronize_iv = 0;
  return;

error:
  must_synchronize_iv = 1;
  return;
}
