#include "list.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* add a char * in the list
 * if e == 0 just add the char *
 * if e == 1 don't modify the list, only put the char *
 * otherwise add the char * followed by e
 * (typical usage: see list_add_dir)
 */
static int _list_add(list *l, char *s, char e)
{
  int len = strlen(s) + 1;

  if (e && e != 1) len++;

  if (e != 1 && l->buf_size + len > l->buf_maxsize) {
    char *new;
    int ns = l->buf_size + len + 4096;
    if (l->buf)
      new = realloc(l->buf, ns);
    else
      new = malloc(ns);
    if (!new) { printf("out of memory\n"); exit(1); }
    l->buf = new;
    l->buf_maxsize = ns;
  }

  if (l->size == l->maxsize) {
    int *new;
    int ns = l->maxsize + 128;
    if (l->data)
      new = realloc(l->data, ns * sizeof(int));
    else
      new = malloc(ns * sizeof(int));
    if (!new) { printf("out of memory\n"); exit(1); }
    l->data = new;
    l->maxsize = ns;
  }

  if (e != 1) {
    l->data[l->size] = l->buf_size;
    l->size++;
  }

  if (e && e != 1) {
    memcpy(l->buf + l->buf_size, s, len-2);
    l->buf[l->buf_size + len - 2] = e;
    l->buf[l->buf_size + len - 1] = 0;
  }
  else
    memcpy(l->buf + l->buf_size, s, len);

  l->buf_size += len;

  if (e != 1 && len - 1 > l->width) l->width = len - 1;

  return 0;
}

int list_add(list *l, char *s)
{
  return _list_add(l, s, 0);
}

int list_add_dir(list *l, char *s)
{
  return _list_add(l, s, '/');
}

int list_add_hidden_int(list *l, int i)
{
  char t[64];
  sprintf(t, "%d", i);
  return _list_add(l, t, 1);
}

int list_size(list *l)
{
  return l->size;
}

char *list_get(list *l, int elem)
{
  return l->buf + l->data[elem];
}

void empty_list(list *h)
{
  h->size = 0;
  h->buf_size = 0;
  h->width = 0;
}

static list *sorted;
static int cmp(const void *_a, const void *_b)
{
  const int *a = _a;
  const int *b = _b;
  char *sa = list_get(sorted, a - sorted->data);
  char *sb = list_get(sorted, b - sorted->data);
  int a_dir, b_dir;
  /* '.' before all */
  if (sa[0] == '.' && sa[1] == '/' && sa[2] == 0) return -1;
  if (sb[0] == '.' && sb[1] == '/' && sb[2] == 0) return 1;
  a_dir = sa[strlen(sa)-1] == '/';
  b_dir = sb[strlen(sb)-1] == '/';
  if (a_dir == b_dir) return strcmp(sa, sb);
  if (a_dir) return -1;
  return 1;
}

void list_sort(list *l)
{
  sorted = l;
  qsort(l->data, l->size, sizeof(int), cmp);
}
