#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifndef WINDOZ
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#else
#include <winsock2.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>

#include "request.h"
#include "command.h"
#include "mess.h"
#include "statistics.h"
#include "iv.h"

char *proxy_dump_file;
char *proxy_dump_file_raw;

extern buf from_server;

static buf data;
static buf out;
static unsigned char reply[4096];
static int reply_size;
static int reply_pos;

static char to_server[2048];
static int to_port;

/* return -1 if error, start of global path on server if ok */
int parse_uri(char *t)
{
  int start_name;
  int end_name;
  int start_port;

  /* port defaults to 80 */
  to_port = 80;

  start_name = 0;
  while (t[start_name] && t[start_name] != '/') start_name++;
  if (!t[start_name]) return -1;
  start_name++; if (t[start_name] != '/') return -1;
  start_name++; if (!t[start_name]) return -1;

  end_name = start_name;
  while (t[end_name] && t[end_name] != '/' && t[end_name] != ':') end_name++;
  if (!t[end_name]) return -1;
  end_name--;
  if (end_name < start_name) return -1;
  if (end_name-start_name+1 > 2047) {
    printf("URI name too long for us: %s\n", t);
    return -1;
  }

  memcpy(to_server, t+start_name, end_name-start_name+1);
  to_server[end_name-start_name+1] = 0;

  if (t[end_name+1] == '/') return end_name+1;

  start_port = end_name+2;

  /* we should be more anal with the port parsing, no? */
  sscanf(t+start_port, "%d", &to_port);
  while (t[start_port] && t[start_port] != '/') start_port++;
  if (!t[start_port]) return -1;

  return start_port;
}

static void get_buffer(int fd)
{
  int d;
#ifndef WINDOZ
  d = read(fd, reply, 4096);
#else
  d = recv(fd, reply, 4096, 0);
#endif
//printf("get_buffer read returns %d\n", d);
  if (d <= 0) return;
  reply_pos = 0;
  reply_size = d;
}

/* return -1 if error */
static int get_char(int fd)
{
  int ret;
  if (reply_pos == reply_size) {
    get_buffer(fd);
    if (reply_pos == reply_size) return -1;
  }
  ret = reply[reply_pos];
  reply_pos++;
  return ret;
}

static void get_line(int fd, buf *b)
{
  int c;

  buf_reset(b);

  while (1) {
    c = get_char(fd);
    /* error or no data, let's pretend it's a end of string */
    if (c == -1) { buf_add_char(b, 0); return; }
    if (c == 13) {
      c = get_char(fd);
      /* error or no data, let's pretend it's a end of string */
      if (c == -1) { buf_add_char(b, 0); return; }
      if (c == 10) { buf_add_char(b, 0); return; }
      buf_add_char(b, 13);
      buf_add_char(b, c);
    }
    buf_add_char(b, c);
  }
  return;
}

static FILE *z;
static void lmess(const char *l, ...)
{
  if (z) {
    va_list ap;
    va_start(ap, l);
    vfprintf(z, l, ap);
    va_end(ap);
    fflush(z);
  } else {
    va_list ap;
    va_start(ap, l);
    vfprintf(stdout, l, ap);
    va_end(ap);
    fflush(stdout);
  }
}

static void lfmess(const void *b, int l)
{
  if (z) {
    fwrite(b, l, 1, z);
    fflush(z);
  } else {
    fwrite(b, l, 1, stdout);
    fflush(stdout);
  }
}

void proxy_mess(void)
{
  mess = lmess;
  fmess = lfmess;
}

int proxy_main(int proxy_port)
{
  int uri;
  int len;
  int i;
  FILE *y;

  if (proxy_dump_file) {
    y = fopen(proxy_dump_file, "wb");
    if (!y) {
      printf("error opening %s\n", proxy_dump_file);
      exit(1);
    }
  }

  if (proxy_dump_file_raw) {
    z = fopen(proxy_dump_file_raw, "wb");
    if (!z) {
      printf("error opening %s\n", proxy_dump_file_raw);
      exit(1);
    }
  }

#ifdef WINDOZ
    WSADATA wsaData;

    int iResult;
    iResult = WSAStartup(MAKEWORD(2, 2), &wsaData);
    if (iResult != 0) {
        printf("WSAStartup failed: %d\n", iResult);
        return 1;
    }
#endif
        struct sockaddr_in addr;
        int fd=socket(PF_INET, SOCK_STREAM, IPPROTO_TCP), fd2;
        unsigned int size=sizeof(struct sockaddr_in);

        if (fd==-1) {
                perror("socket");
                return 1;
        }

#ifndef WINDOZ
	{int adlen=1; setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &adlen, sizeof(int));}
#endif

        addr.sin_family=AF_INET;
        addr.sin_addr.s_addr=INADDR_ANY;
        addr.sin_port=htons(proxy_port);

        if (bind(fd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in))==-1) {
                perror("bind");
                return 1;
        }

next_con:
  reply_size = 0;
  reply_pos = 0;
  buf_reset(&data);
//printf("before listen\n");
        listen(fd, 5);
        fd2=fd;
        if ((fd=accept(fd, (struct sockaddr *)&addr, &size))==-1) {
                perror("accept");
                return 1;
        }
//printf("after accept\n");
//	close(fd2);
	//fprintf(stderr, "connection coming in, let's connect to the other side\n");

  /* get the request */
        while(1) {
	  fd_set rd;

	  FD_ZERO(&rd);
	  FD_SET(fd, &rd);
	  //FD_SET(0, &rd);
	  if (select(fd+1, &rd, 0, 0, 0) == -1) {
	    perror("select");
	    goto fin;
	  }
	  //fprintf(stderr, "select done\n");
	  if (FD_ISSET(fd, &rd)) {
	    //fprintf(stderr, "we have something to read\n");
	    //nb = read(fd, &buf, 4096);
//	    nb = recv(fd, buf, 4096, 0);
  get_line(fd, &data);
//  if (data.size) { printf("got: %s\n", data.buf); }
  if (data.size) break;
	    if (data.size == 0) {
              printf("problem with user's request\n");
	      goto con_over;
	      return 1;
	    }
	  }
        }

  /* blocking some URLs (this should come from a config file) */
  if (!strcmp(data.buf,
              "GET http://pageperso.free.fr/im/css/free.css HTTP/1.1") ||
      !strncmp(data.buf,
              "GET http://pageperso.free.fr/im/free2008/", 41) ||
      !strncmp(data.buf,
              "GET http://safebrowsing-cache.google.com/", 41) ||
      !strncmp(data.buf,
              "POST http://safebrowsing.clients.google.com/", 44)) {
    printf("BLK: %s\n", data.buf);
    goto con_over;
  }

  uri = parse_uri(data.buf);

  printf("REQ: %s\n", data.buf);

  /* get the rest, create request */
  buf_reset(&out);
  /* first line: request */
  /* this for is ugly but should more or less work */
  for (i=0; data.buf[i] && data.buf[i] != ' '; i++)
    buf_add_char(&out, data.buf[i]);
  buf_add_char(&out, ' ');
  buf_add_string(&out, data.buf+uri);
  buf_add_eol(&out);
  /* add Host stuff */
  buf_add_string(&out, "Host: ");
  buf_add_string(&out, to_server);
  if (to_port != 80) { buf_add_char(&out, ':'); buf_add_int(&out, to_port); }
  buf_add_eol(&out);

  len = 0;
  /* get all headers */
  while (1) {
    get_line(fd, &data); if (data.size == 0) goto error;
    if (data.size == 1) break;
    if (!strncmp(data.buf, "Content-Length:", strlen("Content-Length:"))) {
      if (len != 0) goto error;
      if (sscanf(data.buf, "Content-Length: %d", &len) != 1) goto error;
      if (len == 0) goto error;
    } else
    if (!strncmp(data.buf, "Transfer-Encoding: chunked",
                 strlen("Transfer-Encoding: chunked"))) {
      if (len != 0) goto error;
      len = -1;
    } else
    /* "Host: xxx" from client must be removed
     * "Keep-Alive: " as well
     * "Proxy-Connection: " too
     */
    if (!strncmp(data.buf, "Host: ", strlen("Host: ")) ||
        !strncmp(data.buf, "Keep-Alive: ", strlen("Keep-Alive: ")) ||
        !strncmp(data.buf, "Proxy-Connection: ", strlen("Proxy-Connection: ")))
      continue;

    buf_add_string(&out, data.buf);
    buf_add_eol(&out);
  }
  /* force connection: close */
  buf_add_string(&out, "Connection: close"); buf_add_eol(&out);
  buf_add_eol(&out);

  /* get data */
  if (len > 0) {
    while (len) {
      int c = get_char(fd);
      if (c == -1) break;
      buf_add_char(&out, c);
      len--;
    }
  } else {
    /* let's be barbarian and eat up everything we can */
    /* (this may change in the future if something fails */
    fd_set rd;
    while (1) {
      int c;
      struct timeval tm;
      while (1) {
        if (reply_pos == reply_size) break;
        c = get_char(fd);
        if (c == -1) break;
        buf_add_char(&out, c);
      }
      FD_ZERO(&rd);
      FD_SET(fd, &rd);
      tm.tv_sec = 0;
      tm.tv_usec = 0;
      if (select(fd+1, &rd, 0, 0, &tm) == -1) {
        perror("select");
        goto fin;
      }
      if (!FD_ISSET(fd, &rd)) break;
      get_buffer(fd);
      if (reply_pos == reply_size) break;
    }
  }

  if (proxy_dump_file) {
    fprintf(y, "request:*****************************\n");
    fwrite(out.buf, out.size, 1, y);
    fprintf(y, "end request:*************************\n");
    fflush(y);
  }

  /* do the PX command */
  px(&from_server, to_server, to_port, &out);

  printf("  req (s:%d/z:%d/S:%d) rep (s:%d/z:%d/S:%d)\n",
       request_data_size, request_data_zip_size, request_size,
       request_reply_data_size, request_reply_data_zip_size,
       request_reply_size);

  if (from_server.size == 0) goto error;

  if (proxy_dump_file) {
    fprintf(y, "reply:*****************************\n");
    fwrite(from_server.buf, from_server.size, 1, y);
    fprintf(y, "end reply:*************************\n");
    fflush(y);
  }

  printf("REP: ");
  for (i=0; i<from_server.size; i++) {
    if (from_server.buf[i] == '\n' || from_server.buf[i] == '\t') break;
    printf("%c", from_server.buf[i]);
  }
  printf("\n");

  /* send the result back to the requester */
  len = from_server.size;
  i = 0;
  while (len) {
    int d;
#ifndef WINDOZ
    d = write(fd, from_server.buf+i, len);
#else
    d = send(fd, from_server.buf+i, len, 0);
#endif
    if (d <= 0) goto error;
    len -= d;
    i += d;
  }

con_over:
#ifdef WINDOZ
  shutdown(fd, SD_BOTH);
#else
  shutdown(fd, SHUT_RDWR);
#endif
  close(fd);
  fd = fd2;
  goto next_con;

error:
  printf("error\n");
  if (proxy_dump_file) {
    fprintf(y, "error\n");
    fflush(y);
  }
  synchronize_iv(&from_server);
  goto con_over;

fin:
  close(fd);
  close(fd2);
  return 0;
}
